#include "httpd.h"
#include "http_config.h"
#include "http_log.h"

#include "jsapi.h"
#include "jsscript.h"

#include "config.h"

#if CO_DUMPS_JSSTACK
#include "jsdebug.h"
#include "jsprf.h"
/* used with jsdebug functions: */
#define nsnull NULL
#define PRInt32 int32
#endif

#ifdef JS_GCMETER
  #include "jsgc.h"
#endif


#define INSIDE_CO_JS 1

#include "co_lib.h"

#include "co_mod.h"
#include "co_string.h"
#include "co_submodule.h"
#include "co_js.h"
#include "co_jsinheritance.h"
#include "co_jsobjects.h"
#include "ap_parser.h"

#define STACK_CHUNK_SIZE    8192

#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>

/* the "most public" function defined here: retrieves pointer to co_context */
co_context*
cjs_getCOContext(JSContext *cx)
{
    return ((co_js_context*)JS_GetContextPrivate(cx))->cocnx;
}

void
cjs_dumpErrorMessages(JSContext *cx)
{
	co_js_context* cojscnx;
	char *strmsg;
    cojscnx = (co_js_context*)JS_GetContextPrivate(cx);
	strmsg = cos_GetSimplifiedString(cojscnx->cocnx, cojscnx->errmsg);
	cou_sendMessageX( cojscnx->cocnx, strmsg, CO_MSG_FORCE); 
}

void
cjs_setCOInParams(JSContext *cx, void *in_params)
{
    ((co_js_context*)JS_GetContextPrivate(cx))->in_params = in_params;
}

void*
cjs_getCOInParams(JSContext *cx)
{
    return ((co_js_context*)JS_GetContextPrivate(cx))->in_params;
}

void
cjs_setCOMethodName(JSContext *cx, const char *method_name)
{
    ((co_js_context*)JS_GetContextPrivate(cx))->method_name = method_name;
}

const char *
cjs_getCOMethodName(JSContext *cx)
{
    return ((co_js_context*)JS_GetContextPrivate(cx))->method_name;
}

void
cjs_setCOTargetName(JSContext *cx, const char *target_name)
{
    ((co_js_context*)JS_GetContextPrivate(cx))->target_name = target_name;
}

const char *
cjs_getCOTargetName(JSContext *cx)
{
    return ((co_js_context*)JS_GetContextPrivate(cx))->target_name;
}

void
cjs_setCOLineTranslation(JSContext *cx, int (*ln_trans)(const char *, int) )
{
    ((co_js_context*)JS_GetContextPrivate(cx))->ln_trans = ln_trans;
}

int(*
cjs_getCOLineTranslation(JSContext *cx )
)(const char *, int)
{
    return ((co_js_context*)JS_GetContextPrivate(cx))->ln_trans;
}

/* identity for any source of script */
static
int
default_line_translator(const char* source_name, int line)
{
    return line;
}

void
cjs_setDefaultCOLineTranslation(JSContext *cx)
{
    ((co_js_context*)JS_GetContextPrivate(cx))->ln_trans = default_line_translator;
}


static JSClass global_class = {
    "GlobalClass", 0,
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub,   JS_ResolveStub,     JS_ConvertStub,     JS_FinalizeStub,
    JSCLASS_NO_OPTIONAL_MEMBERS
};

static void
co_js_ServerErrorReporter(JSContext *cx, const char *message, JSErrorReport *report)
{
    ap_log_error(APLOG_MARK, CO_LOG_DEBUG, JS_GetContextPrivate(cx),
            "%s", message);
    if (report) {
        if (report->lineno) {
            ap_log_error(APLOG_MARK, CO_LOG_DEBUG, JS_GetContextPrivate(cx),
                "JavaScript error in line: %u", report->lineno);
        }
        if (report->linebuf) {
            ap_log_error(APLOG_MARK, CO_LOG_DEBUG, JS_GetContextPrivate(cx),
                "JavaScript error in text: %s", report->linebuf);
        }
    }
}

static void
co_js_ErrorReporter(JSContext *cx, const char *message, JSErrorReport *report)
{
    co_context *cnx;
	co_string* errmsg;

    errmsg = ((co_js_context*)JS_GetContextPrivate(cx))->errmsg;


//	if (JS_IsExceptionPending(cx)) return;

    cnx = cjs_getCOContext(cx);

    if (report) {
        if (report->lineno) {
			cos_AppendFormatted(cnx, errmsg,
				"JavaScript error: %s in line %d %s %s %s %s",
			    message,
		    	cjs_getCOLineTranslation(cx)(report->filename, report->lineno),
			    (report->filename ? " in " : " "),
			    (report->filename ? report->filename : " "),
			    (report->linebuf ? " in text:\n": " "),
		    	(report->linebuf ? report->linebuf: " ")
			);
        } else if (report->linebuf) {
			cos_AppendFormatted(cnx, errmsg,
				"JavaScript error: %s in text:\n%s",
			    message,
			    report->linebuf
			);
        }
    }
    else {
			cos_AppendString(cnx, errmsg, message); 
    }
#if CO_DUMPS_JSSTACK
{
	char* buf;

//    buf = FormatJSStackDump(cx, nsnull, showArgs, showLocals, showThisProps);
	buf = FormatJSStackDump(cx, nsnull, cjs_getCOLineTranslation(cx), JS_TRUE, JS_FALSE, JS_FALSE);
    if(buf)
    {
		cos_AppendCopyOfString(cnx, errmsg, buf); 
        JS_smprintf_free(buf);
    }
    else
		cos_AppendString(cnx, errmsg, "Failed to format JavaScript stack for dump\n"); 
}
#endif
}


/* set while server process (child) starts */
static
    JSContext *compilation_context = NULL;

static 
const char *
cox_getScript( pool *p, const char *file_name )
{
    int fd;
    const char *content;
    void *map;
    struct stat stat_buf;
    size_t length;

    fd = open( file_name, O_RDONLY);
    if (fd==-1) return NULL; /* can't open file */
    if (fstat(fd, &stat_buf)==-1) return NULL; /* can't stat file */
    length = stat_buf.st_size+1;
    content = ap_pstrdup(p, (map=mmap( 0, length, PROT_READ, MAP_SHARED, fd, 0)));
    munmap(map, length);
    close(fd);
    return content;
}


void
cox_compileJSWrapper( server_rec *s, coffice_jswrapper *jswrapper )
{
	const char* name;
    const char *path;
	const char *bytes;
    JSScript *script;
	pool *local_subpool;
    JSContext *cx;
    JSObject *obj;

	local_subpool = ap_make_sub_pool(cou_getCOModulePool());
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "local_pool size: %ld", ap_bytes_in_pool(local_subpool));

	cx = compilation_context;
	obj = JS_GetGlobalObject(cx);
	name = jswrapper->name;
	path = ap_pstrcat( local_subpool, cou_getCOJsLibPath(), "/co_", name, ".js", NULL);
	ap_log_error(APLOG_MARK, APLOG_DEBUG, s, "mod_coffice: Compiling script: %s for submodule %s", path, name);
	bytes = cox_getScript(local_subpool, path);
	if (bytes) {
	    if ((script=JS_CompileScript( cx, obj, bytes, strlen(bytes), path, 0))==NULL) {
			ap_log_error(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, s, "mod_coffice: Compilation failed");
			jswrapper->script = NULL;
	    }
	    else { /* make object from script for later use */
			jswrapper->script = JS_NewScriptObject( cx, script );
			if (!JS_AddNamedRoot( cx, &(jswrapper->script), name ))
			    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s, "mod_coffice: Rooting script for %s failed", name);
			if (0){
			    jsval retval;
			    JS_CallFunctionName(cx, jswrapper->script, "freeze", 0, NULL, &retval);
			    //ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "frozen script: %s",
			    //	JS_GetStringBytes(JSVAL_TO_STRING(retval)) );
			}
		}
		JS_GC(cx);
	} else jswrapper->script = NULL; /* bytes */
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "local_pool size: %ld", ap_bytes_in_pool(local_subpool));
	ap_destroy_pool(local_subpool);
}

static const char *query_modules =
"
select distinct C.class_name from 
co_sys_meta_method M, co_sys_meta_class C where M.id_meta_class = C.id
 ;
";

void
cox_loadCOModuleScripts(server_rec *s)
{
	coffice_server_conf *coffice_sconf;
	coffice_jswrapper *jswrapper;
    PGresult *pgres;
    int ntuples;
    int i;
    const char *class_name;
	pool *p;

	p = cou_getCOModulePool();

	coffice_sconf = cox2_getCOfficeConfig( s );

    pgres = PQexec(coffice_sconf->dbconn, query_modules);
    if (PQresultStatus(pgres) != PGRES_TUPLES_OK)
    {
        PQclear(pgres);
        return;
    }

    ntuples = PQntuples(pgres);

    for (i=0; i<ntuples; i++) {
        class_name = PQgetvalue(pgres, i, 0);
	    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, 
			"cox_loadCOModules: loading script module %s",	class_name
		);
		/* do it */
///cou_getCOLoadedLibs()
		jswrapper = ap_push_array( cou_getCOLoadedJSWrappers() );
		jswrapper->name = ap_pstrdup( p, class_name);
		cox_compileJSWrapper(s, jswrapper );
    }

    PQclear(pgres);
}

/* |APLOG_NOERRNO added in log_error calls */
void
cox_compileJSScript( server_rec *s )
{
    JSContext *cx;
    JSScript *script;
    JSObject *obj;
    const char *path;
	pool *p;
	coffice_server_conf *coffice_sconf;

	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: Start Building inheritance script" );

	p = cou_getCOModulePool();
	coffice_sconf = cox2_getCOfficeConfig( s );
	//ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: OK1 Building inheritance script" );

    if ((coffice_sconf->jsrt)==NULL) {
	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s, "mod_coffice: Compilation: No JS Runtime! Exiting");
	return;
    }
	; /*failed */
    ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: Creating compilation_context");
    cx = JS_NewContext( coffice_sconf->jsrt, 8192);
    //ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: OK2 compilation_context");
    compilation_context = cx;
    //ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: OK3 compilation_context");
    if (cx==NULL)
	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s, "mod_coffice: Cannot create context");
    //ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: OK4 compilation_context");
    JS_SetErrorReporter(cx, co_js_ServerErrorReporter);
    //ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: OK5 compilation_context");
    JS_SetContextPrivate(cx, s);

    //ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: OK6 compilation_context");
    obj = JS_NewObject(cx, &global_class, NULL, NULL);
    JS_InitStandardClasses(cx, obj);
//    JS_AddRoot(cx, obj); //keep global object
    //ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: OK7 compilation_context");
    if (obj==NULL)
	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s, "mod_coffice: Cannot create global object");

//    JS_GC(cx);

    //ap_log_error(APLOG_MARK, APLOG_DEBUG, s, "mod_coffice: Compilation environment ready");

    {//char* ptrs[2];
	const char **scripts;
	pool *local_subpool;/****/
	local_subpool = ap_make_sub_pool(p);
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "local_pool size: %ld", ap_bytes_in_pool(local_subpool));
	//local_subpool = p;
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: Building inheritance script" );

//	scripts = cox_buildInheritanceScript( s, local_subpool, coffice_sconf->dbconn );
	scripts = cox_buildInheritanceScript( s, p, coffice_sconf->dbconn );

	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: classes: check len");
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: classes: ptr=%x", scripts[0]);
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: classes: len=%d", strlen(scripts[0]));
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: implementations: len=%d", strlen(scripts[1]));

	// XXX
#if 0
	    coffice_sconf->script_classes = NULL;
	    coffice_sconf->script_implementations = NULL;
	    /* should not use JavaScript */
	    return;
#endif
	//ptrs[0] = NULL;
	//ptrs[1] = NULL;
	//scripts = ptrs;// XXX
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: Compiling inheritance script" );
	if ((scripts[0]==NULL)||(scripts[1]==NULL)) {
	    /* something went wrong */
	    coffice_sconf->script_classes = NULL;
	    coffice_sconf->script_implementations = NULL;
	    /* should not use JavaScript */
	    return;
	}

	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: classes: %s", scripts[0]);
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "mod_coffice: implementations: %s", scripts[1]);

	if ((script=JS_CompileScript( cx, obj, scripts[0], strlen(scripts[0]), "inheritance", 0))==NULL) {
	    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s, "mod_coffice: Compilation of classes failed");
	    coffice_sconf->script_classes = NULL;
	}
	else { /* make object from script for later use */
	    coffice_sconf->script_classes = JS_NewScriptObject( cx, script );
	    if (!JS_AddNamedRoot( cx, &(coffice_sconf->script_classes), "classes" ))
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s, "mod_coffice: Rooting classes script failed" );
        }
    JS_GC(cx);
#if 1
	if ((script=JS_CompileScript( cx, obj, scripts[1], strlen(scripts[1]), "inheritance", 0))==NULL) {
	    ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s, "mod_coffice: Compilation of implementations failed");
	    coffice_sconf->script_implementations = NULL;
	}
	else { /* make object from script for later use */
	    coffice_sconf->script_implementations = JS_NewScriptObject( cx, script );
	    if (!JS_AddNamedRoot( cx, &(coffice_sconf->script_implementations), "implementations" ))
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, s, "mod_coffice: Rooting implementations script failed" );
        }
#endif
		ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "local_pool size: %ld", ap_bytes_in_pool(local_subpool));
		ap_destroy_pool(local_subpool);
    }
    JS_GC(cx);
	cox_loadCOModuleScripts(s);

    /* force garbage collection, so that any problems we get right now */
    JS_GC(cx);
    /* we can collect garbage, but we have to keep compilation context
	or JS will core dump while executing compiled script */
    /* TEST TEST */
    //JS_DestroyContext(cx);
}

void
cox_initJSRuntime(coffice_server_conf *coffice_sconf)
{
    JSRuntime *rt;
    /* You need a runtime and one or more contexts to do anything with JS. */
//    rt = JS_NewRuntime(2000000L);
    rt = JS_NewRuntime(100000L);// small ;)
    if (!rt) {
        /*fail("can't create JavaScript runtime");*/
    }
    else 
	coffice_sconf->jsrt = rt;

}

void
cox_finishJSRuntime(coffice_server_conf *coffice_sconf)
{
    JS_DestroyRuntime(coffice_sconf->jsrt);
}

/* create JavaScript context for given request context, using this server's
   JS runtime */
/* if for_asp argument is set to true, line translation is used */
void
cox_initJSContext( co_context *cnx, bool for_asp)
{
    JSContext *cx;
    JSRuntime *rt;

    cou_logError( cnx, CO_LOG_DEBUG, "in initJSContext" );
    rt = cox_getCOfficeConfig(cnx)->jsrt;
    if (rt == NULL) cou_error( cnx, "JavaScript runtime not initialized" );

    cou_logError( cnx, CO_LOG_DEBUG, "creating context..." );
    cx = JS_NewContext( rt, STACK_CHUNK_SIZE);
    if (!cx) {
	cou_error( cnx, "can't create JavaScript context");
    }
    cou_logError( cnx, CO_LOG_DEBUG, "allocating cojscnx..." );
    {
	co_js_context* cojscnx = ap_palloc( cnx->request->pool, sizeof(co_js_context));
	cojscnx->cocnx = cnx;
	cojscnx->errmsg = cos_CreateEmpty(cnx);
	cojscnx->in_params = NULL;
	cojscnx->writes_buffer = NULL;
	cojscnx->buffer_length = 0;
	cojscnx->withinJSCreation = 0;

/*
	if (for_asp)
	    cojscnx->ln_trans = co_translate_line;
	else
*/
	cojscnx->ln_trans = default_line_translator;
	cnx->jscnx = cx;
	cou_logError( cnx, CO_LOG_DEBUG, "setting private data..." );
	JS_SetContextPrivate(cx, cojscnx);
	cou_logError( cnx, CO_LOG_DEBUG, "setting ErrorReporter..." );
	JS_SetErrorReporter(cx, co_js_ErrorReporter);
    }

    /* initialization of JS environment: standard and custom classes/functions */
    {
	JSObject *globalObj;

	//cou_logError( cnx, CO_LOG_DEBUG, "creating global object..." );

	if ((globalObj = JS_NewObject(cx, &global_class, 0, 0))==NULL)
	    cou_error( cnx, "can't create JavaScript global object");
	//JS_AddRoot(cx, globalObj); //keep global object
	//cou_logError( cnx, CO_LOG_DEBUG, "initializing standard classes..." );
	JS_InitStandardClasses(cx, globalObj);

	cjs_InitCOClasses(cx);
    JS_GC(cx);

	//cou_logError( cnx, CO_LOG_DEBUG, "defining functions..." );
	JS_DefineFunctions(cx, globalObj, cjs_functions);
    JS_GC(cx);

    {
	coffice_jswrapper* elts;
	coffice_jswrapper* elt;
	array_header* lib_array;
	const char *name;
	JSObject *script_obj;
	JSScript *script;
	JSBool ok;
	jsval rval;

	int i;


	/* inheritance script */
	script_obj = cox_getCOfficeConfig(cnx)->script_classes;
	if (script_obj) {
		script = (JSScript*)JS_GetPrivate( cx, script_obj );
		cou_logError( cnx, CO_LOG_DEBUG, "executing inheritance script at %0x: data=%0x", (int)script, *(int*)script );
    		ok = JS_ExecuteScript(cx, globalObj, script,
                           &rval);
        	//cou_logError( cnx, CO_LOG_DEBUG, "...executed");
    		if (!ok)
			{
        	    cou_logError( cnx, CO_LOG_DEBUG, "initJSContext: executing precompiled classes script failed!" );
			}
	}
    JS_GC(cx);

	//cou_logError( cnx, CO_LOG_DEBUG, "iterating thru lib_array..." );

	lib_array = cou_getCOLoadedJSWrappers();
	elts = (coffice_jswrapper*)lib_array->elts;
	for (i = 0; i < lib_array->nelts; i++) {
    	    elt = &elts[i];
    	    name = elt->name;
	    script_obj = elt->script;
	    if (script_obj) {
		//cou_logError( cnx, CO_LOG_DEBUG, "script object at %0x for %s: data=%0x", script_obj, name, *(int*)script_obj );
		script = (JSScript*)JS_GetPrivate( cx, script_obj );
		cou_logError( cnx, CO_LOG_DEBUG, "executing script at %0x for %s: data=%0x", (int)script, name, *(int*)script );
		cou_logError( cnx, CO_LOG_DEBUG, "filename=%s", script->filename );
    		ok = JS_ExecuteScript(cx, globalObj, script,
                           &rval);
        	//cou_logError( cnx, CO_LOG_DEBUG, "...executed");
    		if (!ok)
				{
					cou_logError( cnx, CO_LOG_ERR, "initJSContext: executing precompiled script for module %s failed!",name);
				}

    	    } else {cou_logError( cnx, CO_LOG_DEBUG, "no script object for module %s", name);}
    	    //cou_logError( cnx, CO_LOG_DEBUG, "iteration %i finished",i);
	    JS_GC(cx);
	}
	//cou_logError( cnx, CO_LOG_DEBUG, "...end of iterating thru lib_array" );

	/* inheritance script */
	script_obj = cox_getCOfficeConfig(cnx)->script_implementations;
	if (script_obj) {
		script = (JSScript*)JS_GetPrivate( cx, script_obj );
		//cou_logError( cnx, CO_LOG_DEBUG, "executing inheritance script at %0x: data=%0x", (int)script, *(int*)script );
    		ok = JS_ExecuteScript(cx, globalObj, script,
                           &rval);
        	//cou_logError( cnx, CO_LOG_DEBUG, "...executed");
    		if (!ok)
			{
        	    cou_logError( cnx, CO_LOG_ERR, "initJSContext: executing precompiled inheritance script failed!" );
			}
	}
    }
    } /* globalObj visiblility */
	JS_GC(cx);
}

/* destroy all contexts in server's runtime */
void
cox_finishJSContext(co_context *cnx)
{
    JSContext *acx;
    JSContext *iterp = NULL;
//    JSContext *iterp = cnx->jscnx;
    int i=0;
    JSRuntime *rt;
    rt = cox_getCOfficeConfig(cnx)->jsrt;
#ifdef JS_GCMETER
    {
	FILE *file;
	file = fopen("/etc/httpd/logs/js_log", "a");
	if (file) {
	    js_DumpGCStats(rt, file);
	    fclose(file);
	}
	else {
	    cou_logError( cnx, CO_LOG_ERR, "Cannot open js_log for append.");
	}
    }
#endif
    if (rt) {
	cou_logError( cnx, CO_LOG_DEBUG, "destroying all JavaScript contexts...");
	while ( (acx = JS_ContextIterator( rt, &iterp )) != NULL ) {    
		cou_logError( cnx, CO_LOG_DEBUG, "i=%d, acx=%x", i, (int)acx );
	    if (acx!=compilation_context) { 
		JS_GC(acx);
		cou_logError( cnx, CO_LOG_DEBUG, "after GC" );
		//JS_DestroyContext(acx); 
	    }
	    else cou_logError( cnx, CO_LOG_DEBUG, "compilation context skipped" );
	    i++;
        }
	//cou_logError( cnx, CO_LOG_DEBUG, "Nullifying contexts pointer");
	cnx->jscnx = NULL;
	cou_logError( cnx, CO_LOG_DEBUG, "JavaScript contexts destroyed");
    } else
	cou_logError( cnx, CO_LOG_WARNING, "No JavaScript runtime found");
}

